{********************************************************************}
{                                                                    }
{ written by TMS Software                                            }
{            copyright (c) 2018 - 2020                               }
{            Email : info@tmssoftware.com                            }
{            Web : http://www.tmssoftware.com                        }
{                                                                    }
{ The source code is given as is. The author is not responsible      }
{ for any possible damage done due to the use of this code.          }
{ The complete source code remains property of the author and may    }
{ not be distributed, published, given or sold in any form as such.  }
{ No parts of the source code can be included in any other component }
{ or application without written authorization of the author.        }
{********************************************************************}

unit TMSWebCoreIDEOptions;

{$mode objfpc}{$H+}
{$Inline on}

interface

uses
  Classes, SysUtils,
  {$IFDEF Unix}BaseUnix,{$ENDIF}
  {$IFDEF Windows}Registry,{$ENDIF}
  math, Pipes, process, IniFiles, fphttpclient,
  // LazUtils
  LazVersion, LazLoggerBase, LazFileCache, LazConfigStorage, LazFileUtils,
  FileUtil, LazUTF8Classes, LazUTF8, LazStringUtils, Masks,
  // LCL
  Dialogs, Forms, Controls, LCLIntf,
  // IDEIntf
  ProjectIntf, IDEExternToolIntf, MacroDefIntf, MacroIntf, CompOptsIntf,
  BaseIDEIntf, LazIDEIntf, IDEDialogs, IDEMsgIntf, SrcEditorIntf, PackageIntf,
  FormEditingIntf;

Type
  TECMAScript = (esNone,esECMA5,esECMA6);
  TWebServerVisibility = (wsvNone,wsvHidden,wsvVisible);
  TBrowser = (
    bNothing,
    bDefault,
    bSystem,
    bChrome,
    bFirefox,
    {$IFDEF Windows}
    bEdge,
    bIE,
    {$ENDIF}
    bOpera
    // tmswebcompiler supports Firefox developer edition
    );

const
  BrowserName: array[TBrowser] of string = (
    'None',
    'Default',
    'System',
    'Chrome',
    'Firefox',
    {$IFDEF Windows}
    'Edge',
    'IE',
    {$ENDIF}
    'Opera'
    );
  BrowserToWebRunnerID: array[TBrowser] of integer = (
    0, // bNothing => null
    1, //bDefault => default
    2, //bSystem => all
    3, //bChrome
    4, //bFirefox
    {$IFDEF Windows}
    5, //bEdge
    6, //bIE
    {$ENDIF}
    7 //bOpera
    // 8 = FirefoxDeveloperEdition
    );

  // lpk
  TMSWEBCoreRTL_PkgName = 'pas2js_rtl'; // Linux needs case sensitive
  TMSWEBCoreRTL_PkgFile = 'Core Source'+PathDelim+'RTL'+PathDelim+TMSWEBCoreRTL_PkgName+'.lpk';
  TMSWEBCore_PkgName = 'tmswebcore'; // Linux needs case sensitive
  TMSWEBCore_PkgFile = TMSWEBCore_PkgName+'.lpk';
  TMSWEBCoreDe_PkgName = 'tmswebcorepkgdelaz'; // Linux needs case sensitive
  TMSWEBCoreDe_PkgFile = TMSWEBCoreDe_PkgName+'.lpk';
  TMSWEBCoreLib_PkgName = 'tmswebcorepkgliblaz'; // Linux needs case sensitive
  TMSWEBCoreLib_PkgFile = TMSWEBCoreLib_PkgName+'.lpk';

  // keys for project custom data
  KeyTMSWebCoreProject = 'WebCoreProject'; // '1' = true
  KeyProjectRunOnReady = 'RunOnReady'; // '1' = true
  KeySingleJS = 'SingleJS'; // '1' = true
  KeyStartWebServer = 'RunWebServer'; // '1' = true
  KeyRunWebRunner = 'RunWebRunner'; // '1' = true
  // keys for file custom data
  KeyIsProjectHTMLFile = 'IsProjectHTMLFile'; // '1' = true

  // compiler
  {$IFDEF Linux}
  WebCoreBinOS = 'Linux';
  {$ELSE}
    {$IFDEF Darwin}
    WebCoreBinOS = 'macOS';
    {$ELSE}
      {$IFDEF Windows}
      WebCoreBinOS = 'Win32';
      {$ELSE}
      WebCoreBinOS = '$TargetOS()';
      {$ENDIF}
    {$ENDIF}
  {$ENDIF}
  WebCoreDefaultBin = '$(TMSWebDir)'+PathDelim+'Bin'+PathDelim+WebCoreBinOS+PathDelim;
  WebCoreDefaultWebCompiler = WebCoreDefaultBin+'$MakeExe(IDE,TMSWebCompiler)';
  WebCoreDefaultPas2js = '$(TMSWebDir)'+PathDelim+'Compiler'+PathDelim+'$MakeExe(IDE,pas2js)';
  WebCoreDefaultOutputPath = 'html';
  // web server
  WebCoreWebServerCaption = 'Web Server';
  WebCoreDefaultStartWebServer = True;
  WebCoreDefaultWebServerPort = '8000';
  {$IFDEF Unix}
  WebCoreDefaultWebRunner = ''; // empty = use browser
  WebCoreDefaultWebServer = '$(TMSWebDir)'+PathDelim+'Compiler'+PathDelim+'compileserver_$TargetOS(IDE)_$TargetCPU(IDE)';
  WebCoreDefault_URL = 'http://localhost:$(WebCorePort)/$Name($(WebCoreHTML))';
  WebCoreDefaultWebServerParams = '--port=$(WebCorePort) --noindexpage --simpleserver --directory="$Project(OutputDir)"';
  WebCoreDefaultLaunchingAppPlusParams = '"$(WebCoreBrowser)" "$(TMSDefaultURL)"';
  {$ELSE}
  WebCoreDefaultWebRunner = WebCoreDefaultBin+'$MakeExe(IDE,TMSWebRunner)';
  WebCoreDefaultWebServer = WebCoreDefaultBin+'$MakeExe(IDE,TMSWebServerManager)';
  WebCoreDefault_URL = 'http://localhost:$(WebCorePort)/$NameOnly($(ProjFile))/$Name($(WebCoreHTML))';
  WebCoreDefaultWebServerParams = '-s "$(WebCoreDefaultURL)" "$Project(OutputDir)"';
  WebCoreDefaultLaunchingAppPlusParams = '"$(WebCoreRunner)" "$(TMSRunnerIni)"';
  {$ENDIF}
  WebCoreDefaultWebServerVisibility = wsvNone;
  // browser
  WebCoreDefaultRunWebRunner = True;
  WebCoreDefaultBrowser = bDefault;
  WebCoreRunnerIniFile = 'TMSWebRunner.ini';
  // JavaScript
  WebCoreDefaultRunOnReady = False;
  WebCoreDefaultSingleJSFile = True;
  WebCoreDefaultECMAScript = esNone;
  WebCoreDefaultJavascriptLibManager = '$(TMSWebDir)'+PathDelim+'config'+PathDelim+'extensions.ini';
  WebCoreDefaultUserJavascriptLibManager = '';
  WebCoreDefaultComponentUnits = 'WebLib.*';

Const
  WebCoreDsgnOptsFile = 'tmswebcoreoptions.xml';
  // XML paths inside tmswebcoreoptions.xml
  CfgVersion = 1;
  // Version 1: store Browser as name, instead of int
  CfgKeyVersion = 'Version';
  CfgKeyWebInstallDir = 'WebInstallDir/value';
  CfgKeyWebDirAskRename = 'WebInstallDir/askrename';
  CfgKeyWebCompiler = 'WebCompiler/value';
  CfgKeyWebServer = 'WebServer/value';
  CfgKeyWebServerParams = 'WebServer/params/value';
  CfgKeyWebServerPort = 'WebServer/Port';
  CfgKeyWebServerVisibility = 'WebServer/Visibility/value';
  CfgKeyWebRunner = 'WebRunner/value';
  CfgKeyJavascriptLibManager = 'JavascriptLibrary/value';
  CfgKeyUserJavascriptLibManager = 'JavascriptLibrary/User/value';
  CfgKeyURL = 'URL/value';
  CfgKeySingleJSFile = 'SingleJSFile/value';
  CfgKeyBrowser = 'Browser/value';
  CfgKeyComponentUnits = 'ComponentUnits/value';

Type
  // All options containing macro values are cached.
  TWebCoreCachedOption = (
    wccoTMSWebDir,
    wccoWebCompiler, // tmswebcompiler.exe
    wccoURL, // full URL to main html page
    wccoJavascriptLibManager,
    wccoUserJavascriptLibManager,
    wccoWebRunner, // tmswebrunner.exe
    wccoWebServer, // tmswebserver.exe
    wccoWebServerParams,
    wccoWebServerPort,
    wccoComponentUnits
    );
  TWebCoreCachedOptions = set of TWebCoreCachedOption;

  TWCCachedValue = record
    RawValue: string;
    Stamp: int64;
    ParsedValue: string;
  end;
  PPas2jsCachedValue = ^TWCCachedValue;

  { TTMSWebcoreIDEOptions }

  TTMSWebcoreIDEOptions = class
  private
    FCachedOptions: array[TWebCoreCachedOption] of TWCCachedValue;
    FChangeStamp: int64;
    FCheckLicense: boolean;
    FSavedStamp: int64;
    FTMSWebDirAskRename: boolean;
    FWebCoreBrowserMacro: TTransferMacro;
    FWebCoreCompilerMacro: TTransferMacro;
    FWebCoreCompilerOptsMacro : TTransferMacro;
    FWebCoreDefaultURLMacro: TTransferMacro;
    FWebCoreHTMLMacro : TTransferMacro;
    FWebCoreRunnerIniMacro : TTransferMacro;
    FWebCoreRunnerMacro: TTransferMacro;
    FWebCoreServerPortMacro: TTransferMacro;
    FWebCoreURLMacro: TTransferMacro;
    FWebserverVisibility : TWebServerVisibility;
    FSingleJSFile : Boolean;
    FBrowser : TBrowser;
    FTMSWebDirMacro : TTransferMacro;
    Function GetCachedOption(Option: Integer) : String;
    function GetModified: boolean;
    procedure SetBrowser(AValue: TBrowser);
    procedure SetCachedOption(Option: Integer; AValue: string);
    procedure SetModified(AValue: boolean);
    procedure SetSingleJSFile(AValue: Boolean);
    procedure SetWebserverVisibility(AValue: TWebServerVisibility);
  Protected
    function GetTMSWebDirMacro(const {%H-}s: string; const {%H-}Data: PtrInt; var {%H-}Abort: boolean): String;virtual;
    function GetWebCoreDefaultURLMacro(const {%H-}s: string; const {%H-}Data: PtrInt; var {%H-}Abort: boolean): String; virtual;
    function GetWebCoreURLMacro(const {%H-}s: string; const {%H-}Data: PtrInt; var {%H-}Abort: boolean): String; virtual;
    function GetWebCompilerOptionsMacro(const {%H-}s: string; const {%H-}Data: PtrInt; var {%H-}Abort: boolean): string; virtual;
    function GetWebCompilerMacro(const {%H-}s: string; const {%H-}Data: PtrInt; var {%H-}Abort: boolean): String;virtual;
    function GetWebRunnerMacro(const {%H-}s: string; const {%H-}Data: PtrInt; var {%H-}Abort: boolean): string; virtual;
    function GetWebCoreBrowserMacro(const {%H-}s: string; const {%H-}Data: PtrInt; var {%H-}Abort: boolean): string; virtual;
    function GetWebCorePortMacro(const {%H-}s: string; const {%H-}Data: PtrInt; var {%H-}Abort: boolean): string; virtual;
    function GetTMSRunnerIniMacro(const {%H-}s: string; const {%H-}Data: PtrInt; var {%H-}Abort: boolean): string; virtual;
    function GetWebCoreHTMLMacro(const {%H-}s: string; const {%H-}Data: PtrInt; var {%H-}Abort: boolean): string; virtual;
  public
    constructor Create;
    // Reset defaults.
    Procedure Reset; virtual;
    // Get a parsed (= all macros expanded) string configuration variable.
    function GetParsedOptionValue(Option: TWebCoreCachedOption): string; virtual;
    function GetDefaultURL: string; virtual;
    function GetProjectDir(Prj: TLazProject): string; virtual;
    function GetHTMLFile(Prj: TLazProject): TLazProjectFile; virtual;
    function GetHTMLFilename(Prj: TLazProject): string; virtual;
    function GetOutputDir(Prj: TLazProject): String; virtual;
    function GetWebRunnerIniFilename(Prj: TLazProject): String; virtual;
    function GetBrowserPath(URL: string; Params: TStrings): string; virtual;
    function GetBrowserChrome(URL: string; Params: TStrings): string; virtual;
    function GetBrowserFirefox(URL: string; Params: TStrings): string; virtual;
    function GetBrowserOpera(URL: string; Params: TStrings): string; virtual;
    {$IFDEF Windows}
    function GetBrowserEdge(URL: string; Params: TStrings): string; virtual;
    function GetBrowserIE(URL: string; Params: TStrings): string; virtual;
    {$ENDIF}
    function GetBrowserDefault(URL: string; Params: TStrings): string; virtual;
    function FindBrowserFile(ShortFilename: string): string; virtual;
    function FindBrowserPath(Filenames: array of string; URL: string; Params: TStrings): string; virtual;
    function IsUnitAllowed(AnUnitName: string): boolean; virtual;
    function IsComponentAllowed(aComponentClass: TComponentClass): boolean; virtual;
    function CheckTMSWebDir(NewDir: string): boolean; virtual;
    procedure FixTMSWebDir; virtual;
    // Increase change timestamp.
    procedure IncreaseChangeStamp; inline;
    // Create IDE settings storage and call LoadFromConfig
    procedure Load; virtual;
    // Create IDE settings storage and call SaveToFromConfig
    procedure Save; virtual;
    // Load from IDE settings file.
    procedure LoadFromConfig(Cfg: TConfigStorage); virtual;
    // Save to IDE settings file.
    procedure SaveToConfig(Cfg: TConfigStorage); virtual;
  public
    // macros
    property TMSWebDirMacro : TTransferMacro read FTMSWebDirMacro;
    property WebCoreDefaultURLMacro : TTransferMacro read FWebCoreDefaultURLMacro;
    property WebCoreURLMacro : TTransferMacro read FWebCoreURLMacro;
    property WebCoreCompilerMacro : TTransferMacro read FWebCoreCompilerMacro;
    property WebCoreCompilerOptsMacro : TTransferMacro read FWebCoreCompilerOptsMacro;
    property WebCoreHTMLMacro : TTransferMacro read FWebCoreHTMLMacro;
    property WebCoreRunnerMacro : TTransferMacro read FWebCoreRunnerMacro;
    property WebCoreBrowserMacro : TTransferMacro read FWebCoreBrowserMacro;
    property WebCoreWebRunnerIniMacro : TTransferMacro read FWebCoreRunnerIniMacro;
    property WebCoreServerPortMacro : TTransferMacro read FWebCoreServerPortMacro;
    // String values are cached
    property TMSWebDir : string Index Ord(wccoTMSWebDir) read GetCachedOption Write SetCachedOption;
    property TMSWebDirAskRename: boolean read FTMSWebDirAskRename write FTMSWebDirAskRename;
    property WebCompiler : string Index Ord(wccoWebCompiler) read GetCachedOption Write SetCachedOption;
    Property WebCoreURL : String Index Ord(wccoURL) read GetCachedOption Write SetCachedOption;
    Property JavaScriptLibConfig : String Index Ord(wccoJavascriptLibManager) read GetCachedOption Write SetCachedOption;
    Property UserJavaScriptLibConfig : String Index Ord(wccoUserJavascriptLibManager) read GetCachedOption Write SetCachedOption;
    Property WebRunner : String Index Ord(wccoWebRunner) read GetCachedOption Write SetCachedOption;
    Property WebServer : string Index Ord(wccoWebServer) read GetCachedOption Write SetCachedOption;
    Property WebServerParams : string Index Ord(wccoWebServerParams) read GetCachedOption Write SetCachedOption;
    Property WebServerPort : String Index Ord(wccoWebServerPort) read GetCachedOption Write SetCachedOption;
    Property ComponentUnits : String Index Ord(wccoComponentUnits) read GetCachedOption Write SetCachedOption;
    // Non-cached
    Property SingleJSFile : Boolean Read FSingleJSFile Write SetSingleJSFile;
    Property WebserverVisibility : TWebServerVisibility Read FWebserverVisibility Write SetWebserverVisibility;
    Property Browser : TBrowser Read FBrowser Write SetBrowser;
    Property CheckLicense: boolean read FCheckLicense write FCheckLicense;
    // Informational
    property ChangeStamp: int64 read FChangeStamp;
    property Modified: boolean read GetModified write SetModified;
  end;

  { TWebCoreEventSink }

  TWebCoreEventSink = Class(TObject)
  protected
    FLastServerExecutable: string;
    FLastServerWorkDir: string;
    FLastServerCmdLineParams: string;
    procedure DoHeaders(Sender: TObject); virtual;
    function IsWebServerRunning(Prj: TLazProject; TimeoutInMS: integer = 200): boolean; virtual;
    function CreateWebRunnerIni(aFileName : String = ''): boolean; virtual;
    function ExecWebServer(Prj: TLazProject): boolean; virtual;
    function ExecWebRunner(Prj: TLazProject): boolean; virtual;
    function CreateDummyGDBExe(Prj: TLazProject): boolean; virtual;
    function UseWebRunner(Prj: TLazProject): boolean; virtual;
    function CheckLicense(Prj: TLazProject): boolean; virtual;
  Public
    function ProjectBuilding(Sender: TObject): TModalResult; virtual;
    function RunDebugInit(Sender: TObject; var Handled: boolean): TModalResult; virtual;
    function RunWithoutDebug(Sender: TObject; var Handled: boolean): TModalResult; virtual;
    function SaveEditorFile(Sender: TObject; aFile: TLazProjectFile;
      SaveStep: TSaveEditorFileStep; TargetFilename: string): TModalResult; virtual;
    function DesignerAddClicked(ADesigner: TIDesigner;
      {%H-}MouseDownComponent: TComponent; {%H-}Button: TMouseButton; {%H-}Shift: TShiftState;
      {%H-}X, {%H-}Y: Integer; var AComponentClass: TComponentClass;
      var {%H-}NewParent: TComponent): boolean; virtual;
  end;


var
  // Set by unit TMSWebCoreDsnRegister during Register
  WebCoreOptions: TTMSWebcoreIDEOptions = nil;

  // WebFormClass is set by unit TMSWebCoreDsgnRegister
  WebFormClass: TCustomFormClass = nil;
  WebDataModuleClass: TDataModuleClass = nil;

type
  TWCRunToolOption = (
    wcrtoConsole,
    wcrtoNoWait);
  TWCRunToolOptions = set of TWCRunToolOption;

Function WebCoreEventSink : TWebCoreEventSink;
function WebCore_RunTool(const Executable: string; Params: TStringList;
  WorkingDirectory: string; const TheOutput, TheErrors: TStrings;
  out TheExitCode: integer; const Options: TWCRunToolOptions): boolean;
function WebCore_RunIDETool(const Executable, WorkingDirectory: string;
  Params: TStringList;
  ViewCaption, ViewName: string; const Options: TWCRunToolOptions): boolean;
function WebCore_RunMacTerminal(Executable, WorkingDirectory: string; Params: TStringList): boolean;

// Return a standard value
function GetStandardCachedOption(O: TWebCoreCachedOption): string;
function NormalizeTMSDir(Directory: string): string;
function NormalizeTMSFilename(Filename: string): string;

function MaybeQuote(S : String) : String;
function GetCompiledTargetOS: string;
function GetCompiledTargetCPU: string;
function NameToBrowser(s: string): TBrowser;

implementation

Var
  EventSink : TWebCoreEventSink;

function WebCoreEventSink: TWebCoreEventSink;
begin
  If (EventSink = Nil) then
    EventSink:=TWebCoreEventSink.Create;
  Result:=EventSink;
end;

function NormalizeTMSDir(Directory: string): string;
begin
  Result:=ChompPathDelim(ResolveDots(Directory));
end;

function NormalizeTMSFilename(Filename: string): string;
begin
  Result:=ResolveDots(Filename);
end;

function MaybeQuote(S: String): String;
begin
  if Pos(' ',S)=0 then
    Result:=S
  else
    Result:='"'+S+'"';
end;

function GetCompiledTargetOS: string;
begin
  Result:=lowerCase({$I %FPCTARGETOS%});
end;

function GetCompiledTargetCPU: string;
begin
  Result:=lowerCase({$I %FPCTARGETCPU%});
end;

function NameToBrowser(s: string): TBrowser;
var
  b: TBrowser;
begin
  for b in TBrowser do
    if SameText(BrowserName[b],s) then exit(b);
  Result:=bDefault;
end;

// Return a standard value
function GetStandardCachedOption(O: TWebCoreCachedOption): string;
var
  Pkg: TIDEPackage;
begin
  Result:='';
  Case O of
    wccoTMSWebDir:
      begin
      Result:=Sysutils.GetEnvironmentVariable('TMSWebDir');
      if (Result<>'') and (Result[1] in ['''','"']) then
        Result:=AnsiDequotedStr(Result,Result[1]);
      if Result='' then
        begin
        Pkg:=PackageEditingInterface.FindPackageWithName(TMSWEBCoreDe_PkgName);
        if Pkg<>nil then
          begin
          Result:=Pkg.DirectoryExpanded;
          if not FilenameIsAbsolute(Result) then
            Result:='';
          end
        else
          debugln(['Error: [TMSWebCoreIDEOptions.GetStandardCachedOption] package ',TMSWEBCoreDe_PkgName,' not installed!?']);
        end;
      // normalize
      Result:=NormalizeTMSDir(Result);
      end;
    wccoWebCompiler :
      begin
      Result:=WebCoreDefaultWebCompiler;
      if not IDEMacros.SubstituteMacros(Result) then
        Result:='tmswebcompiler'+GetExeExt;
      Result:=NormalizeTMSFilename(Result);
      end;
    wccoWebserver :
      begin
      Result:=WebCoreDefaultWebServer;
      if not IDEMacros.SubstituteMacros(Result) then
        Result:='tmswebserver'+GetExeExt;
      Result:=NormalizeTMSFilename(Result);
      end;
    wccoWebServerPort:
      Result:=WebCoreDefaultWebServerPort;
    wccoWebRunner :
      begin
      Result:=WebCoreDefaultWebRunner;
      if not IDEMacros.SubstituteMacros(Result) then
        Result:='tmswebrunner'+GetExeExt;
      Result:=NormalizeTMSFilename(Result);
      end;
    wccoJavascriptLibManager :
      begin
      Result:=WebCoreDefaultJavascriptLibManager;
      if not IDEMacros.SubstituteMacros(Result) then
        Result:='tmswebrunner'+GetExeExt;
      Result:=NormalizeTMSFilename(Result);
      end;
    wccoURL:
      begin
      Result:=WebCoreDefault_URL;
      if not IDEMacros.SubstituteMacros(Result) then
        Result:='http://localhost:'+WebCoreDefaultWebServerPort+'/project1';
      end;
    wccoComponentUnits:
      Result:=WebCoreDefaultComponentUnits;
  end;
end;

{ TTMSWebcoreIDEOptions }

procedure TTMSWebcoreIDEOptions.SetModified(AValue: boolean);
begin
  if AValue then
    IncreaseChangeStamp
  else
    FSavedStamp:=FChangeStamp;
end;

procedure TTMSWebcoreIDEOptions.SetSingleJSFile(AValue: Boolean);
begin
  If (FSingleJSFile=aValue) then exit;
  FSingleJSFile:=aValue;
  IncreaseChangeStamp;
end;

procedure TTMSWebcoreIDEOptions.SetWebserverVisibility(
  AValue: TWebServerVisibility);
begin
  If (FWebserverVisibility=aValue) then exit;
  FWebserverVisibility:=aValue;
  IncreaseChangeStamp;
end;

function TTMSWebcoreIDEOptions.GetCachedOption(Option: Integer): String;
begin
  Result:=FCachedOptions[TWebCoreCachedOption(Option)].RawValue;
end;

function TTMSWebcoreIDEOptions.GetModified: boolean;
begin
  Result:=FSavedStamp<>FChangeStamp;
end;

function TTMSWebcoreIDEOptions.GetTMSRunnerIniMacro(const s: string;
  const Data: PtrInt; var Abort: boolean): string;
Var
  Dir : string;
  Prj: TLazProject;
begin
  Result:='';
  Prj:=LazarusIDE.ActiveProject;
  if Prj=nil then exit;
  Dir:=GetOutputDir(Prj);
  if Dir='' then
    exit;
  Result:=AppendPathDelim(Dir)+'TMSWebRunner.ini';
  //DebugLn(['TMSRunnerIniMacro: ',Result]);
end;

function TTMSWebcoreIDEOptions.GetWebCoreHTMLMacro(const s: string;
  const Data: PtrInt; var Abort: boolean): string;
begin
  Result:=GetHTMLFilename(LazarusIDE.ActiveProject);
  if Result='' then
    Result:='project1.html';
end;

function TTMSWebcoreIDEOptions.GetWebRunnerMacro(const s: string;
  const Data: PtrInt; var Abort: boolean): string;
begin
  Result:=GetParsedOptionValue(wccoWebRunner);
end;

function TTMSWebcoreIDEOptions.GetWebCoreBrowserMacro(const s: string;
  const Data: PtrInt; var Abort: boolean): string;
var
  Params: TStringListUTF8;
begin
  Params:=TStringListUTF8.Create;
  try
    Result:=WebCoreOptions.GetBrowserPath(s,Params);
    if Result='' then
      Result:='ErrorNoWebCoreBrowserFound';
  finally
    Params.Free;
  end;
end;

function TTMSWebcoreIDEOptions.GetWebCorePortMacro(const s: string;
  const Data: PtrInt; var Abort: boolean): string;
begin
  Result:=GetParsedOptionValue(wccoWebServerPort);
end;

function TTMSWebcoreIDEOptions.GetWebCompilerOptionsMacro(const s: string;
  const Data: PtrInt; var Abort: boolean): string;
Const
{$IFDEF WINDOWS}
  Opt = '/';
  LibExt = '.dll';
{$ELSE}
  Opt = '-';
  {$IFDEF DARWIN}
  LibExt = '.dylib';
  {$ELSE}
  LibExt = '.so';
  {$ENDIF}
{$ENDIF}

Var
  SingleJS,
  Optimize,
  DebugInfo : Boolean;
  UseECMAScript,
  CompilerBin,
  IncSearchPaths,
  UnitSearchPaths,
  IntermediateDir,
  ProjectDir,
  HTMLOutputDir , TargetProc, LPIFilename, HTMLProjectFile: String;
  Prj: TLazProject;
  CompOpts: TLazCompilerOptions;
begin
  Prj:=LazarusIDE.ActiveProject;
  if Prj=nil then exit('');
  // Collect all options
  LPIFilename:=Prj.ProjectInfoFile;
  if LPIFilename='' then
    LPIFilename:=AppendPathDelim(LazarusIDE.GetTestBuildDirectory)+'project1.lpi';
  ProjectDir:=GetProjectDir(Prj);

  CompOpts:=Prj.LazCompilerOptions;
  // Check IDE compiler options
  DebugInfo:=CompOpts.GenerateDebugInfo;
  Optimize:=CompOpts.OptimizationLevel>0;
  TargetProc:=CompOpts.TargetProcessor;
  if SameText(TargetProc,'ECMAScript6') then
    UseEcmaScript:='1'
  else
    UseEcmaScript:='0';
  // Other options
  SingleJS:=SingleJSFile; // default value
  if Prj.CustomData.Contains(KeySingleJS) then
    SingleJS:=Prj.CustomData[KeySingleJS]='1'; // project has own value
  IntermediateDir:=ProjectDir+'Intermediate';
  HTMLOutputDir:=CompOpts.GetUnitOutputDirectory(true);
  // Search project HTML file.
  HTMLProjectFile:=GetHTMLFilename(Prj);
  CompilerBin:=AppendPathDelim(GetParsedOptionValue(wccoTMSWebDir))
                       +'Compiler'+PathDelim+'libpas2js'+LibExt;
  IncSearchPaths:=CompOpts.GetIncludePath(true,coptParsed,false);
  UnitSearchPaths:=CompOpts.GetUnitPath(true,coptParsed,false);

  // Now Build command-line options
  Result:=Opt+'HtmlOutputDir:'+MaybeQuote(HTMLOutputDir);
  Result:=Result+' '+Opt+'CompilerBin:'+MaybeQuote(CompilerBin);
  if DebugInfo then
    Result:=Result+' '+Opt+'Debug';
  if SingleJS then
    Result:=Result+' '+Opt+'SingleJS';
  Result:=Result+' '+Opt+'EcmaScript:'+UseECMAScript;
  if IncSearchPaths<>'' then
    Result:=Result+' '+Opt+'IncSearchPaths:'+MaybeQuote(IncSearchPaths);
  if (UnitSearchPaths<>'') then
    Result:=Result+' '+Opt+'UnitSearchPaths:'+MaybeQuote(UnitSearchPaths);
  Result:=Result+' '+Opt+'Optimization:'+IntToStr(Ord(Optimize));
  Result:=Result+' '+Opt+'ProjectHTMLFile:'+MaybeQuote(HTMLProjectFile);
  Result:=Result+' '+Opt+'IntermediateOutputDir:'+MaybeQuote(IntermediateDir);

  if CompOpts.ShowAll then
    Result:=Result+' '+Opt+'ShowEverything';
  if CompOpts.ShowCond then
    Result:=Result+' '+Opt+'ShowConditionals';
  if CompOpts.ShowDebugInfo then
    Result:=Result+' '+Opt+'ShowDebugNotes';
  if CompOpts.ShowHints then
    Result:=Result+' '+Opt+'ShowHints';
  if CompOpts.ShowLineNum then
    Result:=Result+' '+Opt+'ShowLineNumbers';
  if CompOpts.ShowNotes then
    Result:=Result+' '+Opt+'ShowNotes';
  //ShowNothing, example /ShowNothing
  if CompOpts.ShowTriedFiles then
    Result:=Result+' '+Opt+'ShowTriedFiles';
  if CompOpts.ShowExecInfo then
    Result:=Result+' '+Opt+'ShowUsedTools';
  if CompOpts.ShowWarn then
    Result:=Result+' '+Opt+'ShowWarnings';
  // always show full file paths, infos and message numbers
  Result:=Result+' '+Opt+'ShowFileNames '+Opt+'ShowInfo '+Opt+'ShowMessageNumbers';

  // ToDo: MessageFlags to /HiddenMessages:123,456 ...

  //ShowMessageNumbers, example /ShowMessageNumbers

  Result:=Result+' '+Opt+'ProjectFile:'+MaybeQuote(ChangeFileExt(LPIFilename,'.lpr'));
end;

procedure TTMSWebcoreIDEOptions.SetBrowser(AValue: TBrowser);
begin
  if (FBrowser=aValue) then exit;
  FBrowser:=AValue;
  IncreaseChangeStamp;
end;

constructor TTMSWebcoreIDEOptions.Create;
begin
  FChangeStamp:=LUInvalidChangeStamp64;
  Reset;
  FTMSWebDirMacro:=TTransferMacro.Create('TMSWebDir','','TMS Web installation directory',@GetTMSWebDirMacro,[]);
  FWebCoreDefaultURLMacro:=TTransferMacro.Create('WebCoreDefaultURL','','URL directory of main html page',@GetWebCoreDefaultURLMacro,[]);
  FWebCoreURLMacro:=TTransferMacro.Create('WebCoreURL','','URL of main html page',@GetWebCoreURLMacro,[]);
  FWebCoreCompilerMacro:=TTransferMacro.Create('WebCoreCompiler','','Default TMS Web compiler',@GetWebCompilerMacro,[]);
  FWebCoreRunnerMacro:=TTransferMacro.Create('WebCoreRunner','','TMS Web Runner',@GetWebRunnerMacro,[]);
  FWebCoreBrowserMacro:=TTransferMacro.Create('WebCoreBrowser','','Browser path',@GetWebCoreBrowserMacro,[]);
  FWebCoreRunnerIniMacro:=TTransferMacro.Create('TMSRunnerIni','','TMS Web Runner ini file',@GetTMSRunnerIniMacro,[]);
  FWebCoreServerPortMacro:=TTransferMacro.Create('WebCorePort','','TMS Web Server Port',@GetWebCorePortMacro,[]);
  FWebCoreCompilerOptsMacro:=TTransferMacro.Create('WebCoreCompilerOpts','','Default TMS Web compiler options',@GetWebCompilerOptionsMacro,[]);
  FWebCoreHTMLMacro:=TTransferMacro.Create('WebCoreHTML','','TMS Web project html file',@GetWebCoreHTMLMacro,[]);
  // If you ad a macro here, do not forget tmswebcoredsgnregister.pp
end;

function TTMSWebcoreIDEOptions.GetWebCompilerMacro(const s: string; const Data: PtrInt; var Abort: boolean): String;
begin
  Result:=MaybeQuote(WebCompiler);
end;

function TTMSWebcoreIDEOptions.GetTMSWebDirMacro(const s: string; const Data: PtrInt; var Abort: boolean) :  String;
begin
  Result:=TMSWebDir;
end;

function TTMSWebcoreIDEOptions.GetWebCoreDefaultURLMacro(const s: string;
  const Data: PtrInt; var Abort: boolean): String;
begin
  Result:=GetDefaultURL;
end;

function TTMSWebcoreIDEOptions.GetWebCoreURLMacro(const s: string; const Data: PtrInt; var Abort: boolean) :  String;
begin
  Result:=WebCoreURL;
end;

procedure TTMSWebcoreIDEOptions.Reset;
Var
  W : TWebCoreCachedOption;
begin
  FCachedOptions[wccoTMSWebDir].RawValue:=GetStandardCachedOption(wccoTMSWebDir);
  if TMSWebDir='' then
    debugln(['Warning: [TTMSWebcoreIDEOptions.Reset] environment var TMSWebDir and package "'+TMSWEBCoreDe_PkgName+'" is missing']);
  FCachedOptions[wccoWebCompiler].RawValue:=WebCoreDefaultWebCompiler;
  FCachedOptions[wccoWebServer].RawValue:=WebCoreDefaultWebServer;
  FCachedOptions[wccoWebserverParams].RawValue:=WebCoreDefaultWebServerParams;
  FCachedOptions[wccoWebServerPort].RawValue:=WebCoreDefaultWebServerPort;
  FCachedOptions[wccoWebRunner].RawValue:=WebCoreDefaultWebRunner;
  FCachedOptions[wccoJavascriptLibManager].RawValue:=WebCoreDefaultJavascriptLibManager;
  FCachedOptions[wccoUserJavascriptLibManager].RawValue:=WebCoreDefaultUserJavascriptLibManager;
  FCachedOptions[wccoURL].RawValue:=WebCoreDefault_URL;
  FCachedOptions[wccoComponentUnits].RawValue:=WebCoreDefaultComponentUnits;
  for W in TWebCoreCachedOption do
    FCachedOptions[W].Stamp:=LUInvalidChangeStamp64;
  FCheckLicense:=true;
end;

procedure TTMSWebcoreIDEOptions.IncreaseChangeStamp;
begin
  LUIncreaseChangeStamp64(FChangeStamp);
end;

procedure TTMSWebcoreIDEOptions.Load;
var
  Cfg: TConfigStorage;
begin
  Cfg:=GetIDEConfigStorage(WebCoreDsgnOptsFile,true);
  try
    LoadFromConfig(Cfg);
  finally
    Cfg.Free;
  end;
end;

procedure TTMSWebcoreIDEOptions.Save;
var
  Cfg: TConfigStorage;
begin
  Cfg:=GetIDEConfigStorage(WebCoreDsgnOptsFile,false);
  try
    SaveToConfig(Cfg);
  finally
    Cfg.Free;
  end;
end;

procedure TTMSWebcoreIDEOptions.LoadFromConfig(Cfg: TConfigStorage);
var
  FileVersion: Integer;
begin
  With Cfg do
    begin
    FileVersion:=GetValue(CfgKeyVersion,0);
    TMSWebDir:=NormalizeTMSDir(GetValue(CfgKeyWebInstallDir,GetStandardCachedOption(wccoTMSWebDir)));
    if TMSWebDir='' then
      TMSWebDir:=GetStandardCachedOption(wccoTMSWebDir);
    // read TMSWebDirAskRename after TMSWebdDir!
    TMSWebDirAskRename:=GetValue(CfgKeyWebDirAskRename,true);
    WebCompiler:=NormalizeTMSFilename(GetValue(CfgKeyWebCompiler,WebCoreDefaultWebCompiler));
    WebCoreURL:=GetValue(CfgKeyURL,WebCoreDefault_URL);
    JavaScriptLibConfig:=NormalizeTMSFilename(GetValue(CfgKeyJavascriptLibManager,WebCoreDefaultJavascriptLibManager));
    UserJavaScriptLibConfig:=NormalizeTMSFilename(GetValue(CfgKeyUserJavascriptLibManager,WebCoreDefaultUserJavascriptLibManager));
    WebRunner:=NormalizeTMSFilename(GetValue(CfgKeyWebRunner,WebCoreDefaultWebRunner));
    WebServer:=NormalizeTMSFilename(GetValue(CfgKeyWebServer,WebCoreDefaultWebServer));
    WebServerPort:=GetValue(CfgKeyWebServerPort,WebCoreDefaultWebServerPort);
    WebServerParams:=GetValue(CfgKeyWebServerParams,WebCoreDefaultWebServerParams);
    SingleJSFile:=GetValue(CfgKeySingleJSFile,WebCoreDefaultSingleJSFile);
    if FileVersion<1 then
      Browser:=WebCoreDefaultBrowser
    else
      Browser:=NameToBrowser(GetValue(CfgKeyBrowser,BrowserName[WebCoreDefaultBrowser]));
    WebServerVisibility:=TWebServerVisibility(GetValue(CfgKeyWebServerVisibility,Ord(WebCoreDefaultWebServerVisibility)));
    ComponentUnits:=GetValue(CfgKeyComponentUnits,WebCoreDefaultComponentUnits);
    end;
  Modified:=false;
end;

procedure TTMSWebcoreIDEOptions.SaveToConfig(Cfg: TConfigStorage);
begin
  With Cfg do
    begin
    SetValue(CfgKeyVersion,CfgVersion);
    SetDeleteValue(CfgKeyWebInstallDir,TMSWebDir,GetStandardCachedOption(wccoTMSWebDir));
    SetDeleteValue(CfgKeyWebDirAskRename,TMSWebDirAskRename,true);
    SetDeleteValue(CfgKeyWebCompiler,WebCompiler,WebCoreDefaultWebCompiler);
    SetDeleteValue(CfgKeyURL,WebCoreURL,WebCoreDefault_URL);
    SetDeleteValue(CfgKeyJavascriptLibManager,JavaScriptLibConfig,WebCoreDefaultJavascriptLibManager);
    SetDeleteValue(CfgKeyUserJavascriptLibManager,UserJavaScriptLibConfig,WebCoreDefaultUserJavascriptLibManager);
    SetDeleteValue(CfgKeyWebRunner,WebRunner,WebCoreDefaultWebRunner);
    SetDeleteValue(CfgKeyWebServer,WebServer,WebCoreDefaultWebServer);
    SetDeleteValue(CfgKeyWebServerPort,WebServerPort,WebCoreDefaultWebServerPort);
    SetDeleteValue(CfgKeyWebServerParams,WebServerParams,WebCoreDefaultWebServerParams);
    SetDeleteValue(CfgKeySingleJSFile,SingleJSFile,WebCoreDefaultSingleJSFile);
    SetDeleteValue(CfgKeyBrowser,BrowserName[Browser],BrowserName[WebCoreDefaultBrowser]);
    SetDeleteValue(CfgKeyWebServerVisibility,Ord(WebServerVisibility),Ord(WebCoreDefaultWebServerVisibility));
    SetDeleteValue(CfgKeyComponentUnits,ComponentUnits,WebCoreDefaultComponentUnits);
    end;
  Modified:=false;
end;

function TTMSWebcoreIDEOptions.GetProjectDir(Prj: TLazProject): string;
begin
  Result:='';
  if Prj<>nil then
    Result:=Prj.Directory;
  if Result='' then
    ResulT:=LazarusIDE.GetTestBuildDirectory;
end;

function TTMSWebcoreIDEOptions.GetHTMLFile(Prj: TLazProject): TLazProjectFile;
var
  i: Integer;
begin
  if Prj=nil then exit(nil);
  for i:=0 to Prj.FileCount-1 do
    if Prj.Files[i].CustomData[KeyIsProjectHTMLFile]='1' then
      exit(Prj.Files[i]);
  Result:=nil;
end;

function TTMSWebcoreIDEOptions.GetHTMLFilename(Prj: TLazProject): string;
var
  PrjFile: TLazProjectFile;
  LPIFilename: String;
begin
  Result:='';
  PrjFile:=GetHTMLFile(Prj);
  if PrjFile<>nil then
    begin
    Result:=PrjFile.Filename;
    if Result<>'' then
      begin
      if not FilenameIsAbsolute(Result) then
        Result:=ResolveDots(AppendPathDelim(GetOutputDir(Prj))+Result);
      exit;
      end;
    end;
  // If not found, use default: project file, extension .html
  if Prj=nil then exit;
  LPIFilename:=Prj.ProjectInfoFile;
  if LPIFilename='' then exit;
  Result:=ChangeFileExt(LPIFilename,'.html');
  if not FilenameIsAbsolute(Result) then
    Result:=ResolveDots(AppendPathDelim(GetOutputDir(Prj))+Result);
end;

function TTMSWebcoreIDEOptions.GetOutputDir(Prj: TLazProject): String;
begin
  if Prj=nil then exit('');
  Result:=Prj.LazCompilerOptions.GetUnitOutputDirectory(false);
  if Result='' then // Project was not yet saved, take the temporary
    Result:=LazarusIDE.GetTestBuildDirectory;
end;

function TTMSWebcoreIDEOptions.GetWebRunnerIniFilename(Prj: TLazProject
  ): String;
var
  OutputDir: String;
begin
  Result:='';
  if Prj=nil then exit;
  OutputDir:=GetOutputDir(Prj);
  if OutputDir='' then exit;
  Result:=AppendPathDelim(OutputDir)+WebCoreRunnerIniFile;
end;

function TTMSWebcoreIDEOptions.GetBrowserPath(URL: string; Params: TStrings
  ): string;
begin
  case Browser of
    bDefault: Result:=GetBrowserDefault(URL,Params);
    bChrome: Result:=GetBrowserChrome(URL,Params);
    bFirefox: Result:=GetBrowserFirefox(URL,Params);
    {$IFDEF Windows}
    bEdge: Result:=GetBrowserEdge(URL,Params);
    bIE: Result:=GetBrowserIE(URL,Params);
    {$ENDIF}
    bOpera: Result:=GetBrowserOpera(URL,Params);
  else exit('');
  end;
end;

function TTMSWebcoreIDEOptions.GetBrowserChrome(URL: string; Params: TStrings
  ): string;
begin
  Result := FindBrowserPath([
    {$IFDEF Darwin}'/Applications/Google Chrome.app/Contents/MacOS/Google Chrome',{$ENDIF}
    'google-chrome'],URL,Params);
end;

function TTMSWebcoreIDEOptions.GetBrowserFirefox(URL: string; Params: TStrings
  ): string;
begin
  Result := FindBrowserPath([
    {$IFDEF Darwin}'/Applications/Firefox.app/Contents/MacOS/firefox',{$ENDIF}
    'firefox','mozilla'],URL,Params);
end;

function TTMSWebcoreIDEOptions.GetBrowserOpera(URL: string; Params: TStrings
  ): string;
begin
  Result := FindBrowserPath(['opera'],URL,Params);
end;

{$IFDEF Windows}
function TTMSWebcoreIDEOptions.GetBrowserEdge(URL: string; Params: TStrings
  ): string;
begin
  Result := FindBrowserPath(['edge'],URL,Params);
end;

function TTMSWebcoreIDEOptions.GetBrowserIE(URL: string; Params: TStrings
  ): string;
begin
  Result := FindBrowserPath(['iexplore'],URL,Params);
end;
{$ENDIF}

function TTMSWebcoreIDEOptions.GetBrowserDefault(URL: string; Params: TStrings
  ): string;
begin
  Result:=GetBrowserChrome(URL,Params);
  if Result<>'' then exit;
  Result:=GetBrowserFirefox(URL,Params);
  if Result<>'' then exit;
  {$IFDEF Windows}
  Result:=GetBrowserIE(URL,Params);
  if Result<>'' then exit;
  {$ENDIF}
  Result:=GetBrowserOpera(URL,Params);
end;

function TTMSWebcoreIDEOptions.FindBrowserFile(ShortFilename: string): string;
begin
  Result := SearchFileInPath(ShortFilename + GetExeExt, '',
                    GetEnvironmentVariableUTF8('PATH'), PathSeparator,
                    [sffDontSearchInBasePath]);
end;

function TTMSWebcoreIDEOptions.FindBrowserPath(Filenames: array of string;
  URL: string; Params: TStrings): string;
var
  i: Integer;
  aFilename: String;
begin
  Result:='';
  for i:=low(Filenames) to high(Filenames) do
  begin
    aFilename:=Filenames[i];
    if FilenameIsAbsolute(aFilename) then
    begin
      if FileExistsCached(aFilename) then
      begin
        Result:=aFilename;
        break;
      end;
    end else begin
      Result := FindBrowserFile(aFilename);
      if Result<>'' then break;
    end;
  end;
  if Result<>'' then
  begin
    Params.Add(URL);
  end;
end;

function TTMSWebcoreIDEOptions.IsUnitAllowed(AnUnitName: string): boolean;
var
  List, Mask: String;
  MaskItem: TMask;
  p: Integer;
begin
  List:=ComponentUnits;
  //debugln(['TTMSWebcoreIDEOptions.IsUnitAllowed AnUnitName="',AnUnitName,'" List="',List,'"']);
  p:=1;
  repeat
    Mask:=GetNextDelimitedItem(List,';',p);
    if Mask='' then break;
    MaskItem := TMask.Create(Mask{$IF laz_fullversion > 2010000}, []{$ENDIF});
    try
      if MaskItem.Matches(AnUnitName) then
        exit(true);
    finally
      MaskItem.Free;
    end;
  until false;
  Result:=false;
end;

function TTMSWebcoreIDEOptions.IsComponentAllowed(
  aComponentClass: TComponentClass): boolean;
var
  aClass: TClass;
begin
  Result:=true;
  if aComponentClass.InheritsFrom(TControl) then
    begin
    // visual components must descend from weblib units
    aClass:=aComponentClass;
    while aClass<>nil do
      begin
      if IsUnitAllowed(aClass.UnitName) then
        exit(true);
      aClass:=aClass.ClassParent;
      end;
    Result:=false;
    end;
end;

function TTMSWebcoreIDEOptions.CheckTMSWebDir(NewDir: string): boolean;
var
  TMSDir: string;

  function CheckLPK(PkgName, PkgFile: string): boolean;
  var
    MsgResult: Integer;
  begin
    PkgFile:=AppendPathDelim(TMSDir)+PkgFile;
    if FileExistsUTF8(PkgFile) then exit(true);
    MsgResult:=IDEMessageDialog('Error',
      'TMS package "'+PkgName+'" not found:'+sLineBreak
      +'"'+PkgFile+'"',
      mtError,[mbCancel,mbIgnore]);
    Result:=MsgResult=mrIgnore;
  end;

var
  MsgResult: Integer;
begin
  Result:=false;

  TMSDir:=ChompPathDelim(ResolveDots(NewDir));
  if not DirPathExists(TMSDir) then
    begin
    MsgResult:=IDEMessageDialog('Error',
      'TMSWebDir not found: "'+TMSDir+'"',
      mtError,[mbCancel,mbIgnore]);
    exit(MsgResult=mrIgnore);
    end;

  if not CheckLPK(TMSWEBCoreRTL_PkgName,TMSWEBCoreRTL_PkgFile) then exit;
  if not CheckLPK(TMSWEBCore_PkgName,TMSWEBCore_PkgFile) then exit;
  if not CheckLPK(TMSWEBCoreDe_PkgName,TMSWEBCoreDe_PkgFile) then exit;
  if not CheckLPK(TMSWEBCoreLib_PkgName,TMSWEBCoreLib_PkgFile) then exit;

  Result:=true;
end;

procedure TTMSWebcoreIDEOptions.FixTMSWebDir;
{$IFDEF Unix}
var
  CompilerFilename, RealCompilerFile, RealCompilerRelFile,
    WebServerRelFile, WebServerFile: String;
{$ENDIF}
begin
  {$IFDEF Unix}
  CompilerFilename:=WebCoreDefaultPas2js;
  if IDEMacros.SubstituteMacros(CompilerFilename) then
  begin
    if FilenameIsAbsolute(CompilerFilename)
        and (ExtractFilename(CompilerFilename)='pas2js')
        and DirectoryExistsUTF8(ExtractFilePath(CompilerFilename)) then
    begin
      // valid default path to pas2js
      if not FileExistsCached(CompilerFilename) then
      begin
        // missing pas2js executable
        RealCompilerRelFile:='pas2js_'+GetCompiledTargetOS+'_'+GetCompiledTargetCPU;
        RealCompilerFile:=ExtractFilePath(CompilerFilename)+RealCompilerRelFile;
        if FileExistsCached(RealCompilerFile) then
        begin
          if not FileIsExecutable(RealCompilerFile) then
          begin
            // fix executable perms on pas2js_OS_CPU
            FpChmod(RealCompilerFile,&775);
            InvalidateFileStateCache(RealCompilerFile);
          end;
          // symlink pas2js -> pas2js_OS_CPU
          fpSymlink(PChar(RealCompilerRelFile),PChar(CompilerFilename));
          InvalidateFileStateCache(CompilerFilename);
        end;
      end;

      // simple webserver
      WebServerRelFile:='compileserver_'+GetCompiledTargetOS+'_'+GetCompiledTargetCPU;
      WebServerFile:=ExtractFilePath(CompilerFilename)+WebServerRelFile;
      if FileExistsCached(WebServerFile)
          and not FileIsExecutableCached(WebServerFile) then
      begin
        // fix executable perms on compileserver_OS_CPU
        FpChmod(WebServerFile,&775);
        InvalidateFileStateCache(WebServerFile);
      end;
    end;
  end;

  CompilerFilename:=WebCoreDefaultWebCompiler;
  if IDEMacros.SubstituteMacros(CompilerFilename) then
  begin
    if FilenameIsAbsolute(CompilerFilename)
        and SameText(ExtractFilename(CompilerFilename),'tmswebcompiler')
        and DirectoryExistsUTF8(ExtractFilePath(CompilerFilename)) then
    begin
      // valid default path to TMSWebCompiler
      if not FileExistsCached(CompilerFilename) then
      begin
        // missing TMSWebCompiler executable
        RealCompilerRelFile:='TMSWebCompiler_'+GetCompiledTargetOS+'_'+GetCompiledTargetCPU;
        RealCompilerFile:=ExtractFilePath(CompilerFilename)+RealCompilerRelFile;
        if FileExistsCached(RealCompilerFile) then
        begin
          if not FileIsExecutable(RealCompilerFile) then
          begin
            // fix executable perms on TMSWebCompiler_OS_CPU
            FpChmod(RealCompilerFile,&775);
            InvalidateFileStateCache(RealCompilerFile);
          end;
          // symlink TMSWebCompiler -> TMSWebCompiler_OS_CPU
          fpSymlink(PChar(RealCompilerRelFile),PChar(CompilerFilename));
          InvalidateFileStateCache(CompilerFilename);
        end;
      end;
    end;
  end;
  {$ENDIF}
end;

function TTMSWebcoreIDEOptions.GetParsedOptionValue(Option: TWebCoreCachedOption): string;
Const
  FileNames = [wccoTMSWebDir, wccoWebCompiler,
               wccoJavascriptLibManager, wccoUserJavascriptLibManager,
               wccoWebRunner, wccoWebServer];
  Executables = [wccoWebCompiler,wccoWebServer,wccoWebRunner];
var
  p: PPas2jsCachedValue;
begin
  Result:='';
  p:=@FCachedOptions[Option];
  if (p^.Stamp<>IDEMacros.BaseTimeStamp)
      or (laz_fullversion<2000100) // Lazarus 2.0.0 did not increase IDEMacros.BaseTimeStamp
      then
    begin
    p^.Stamp:=IDEMacros.BaseTimeStamp;
    p^.ParsedValue:=p^.RawValue;
    IDEMacros.SubstituteMacros(p^.ParsedValue);
    // Writeln('GetParsedOptionValue: ',Option,' Raw: ',p^.RawValue,' -> ',p^.ParsedValue);
    if Option in filenames then
      begin
      if Option=wccoTMSWebDir then
        p^.ParsedValue:=NormalizeTMSDir(p^.ParsedValue)
      else
        p^.ParsedValue:=TrimFilename(p^.ParsedValue);
      if (p^.ParsedValue<>'') and not FilenameIsAbsolute(p^.ParsedValue) then
        begin
        if (ExtractFilePath(p^.ParsedValue)='') and (Option in Executables) then
          p^.ParsedValue:=FindDefaultExecutablePath(p^.ParsedValue)
        else
          p^.ParsedValue:=''; // not found
        end;
      end;
    if p^.ParsedValue='' then
      p^.ParsedValue:=GetStandardCachedOption(Option);
    end;
  Result:=p^.ParsedValue;
  // Writeln('GetParsedOptionValue: ',Option,' Src: ',GetCachedOption(Ord(Option)),' -> ',Result);
end;

function TTMSWebcoreIDEOptions.GetDefaultURL: string;
// WebCoreDefaultURL is the path of WebCoreURL, e.g. without the last filename
var
  p: Integer;
begin
  Result:=WebCoreOptions.GetParsedOptionValue(wccoURL);
  p:=length(Result);
  while (p>0) and (Result[p]<>'/') do dec(p);
  if p<0 then exit;
  if (p>2) and (Result[p-1]='/') and (Result[p-2]=':') then exit;
  Result:=LeftStr(Result,p-1);
end;

procedure TTMSWebcoreIDEOptions.SetCachedOption(Option: Integer; AValue: string
  );
var
  Opt: TWebCoreCachedOption absolute Option;
begin
  if Opt=wccoTMSWebDir then
    AValue:=NormalizeTMSDir(AValue);
  With FCachedOptions[Opt] do
    begin
    if RawValue=AValue then exit;
    RawValue:=AValue;
    Stamp:=InvalidIDEMacroStamp;
    end;
  if Opt=wccoTMSWebDir then
    TMSWebDirAskRename:=true;
  IncreaseChangeStamp;
  IDEMacros.IncreaseBaseStamp;
end;

{ TWebCoreEventSink }

function WebCore_RunTool(const Executable: string; Params: TStringList;
  WorkingDirectory: string; const TheOutput, TheErrors: TStrings; out
  TheExitCode: integer; const Options: TWCRunToolOptions): boolean;
var
  Buf: string;

  function ReadInputPipe(aStream: TInputPipeStream; var LineBuf: string;
    IsStdErr: boolean): boolean;
  // true if some bytes have been read
  var
    Count: DWord;
    StartPos: Integer;
    i: DWord;
  begin
    Result:=false;
    if aStream=nil then exit;
    Count:=aStream.NumBytesAvailable;
    if Count=0 then exit;
    Count:=aStream.Read(Buf[1],min(length(Buf),Count));
    if Count=0 then exit;
    Result:=true;
    StartPos:=1;
    i:=1;
    while i<=Count do begin
      if Buf[i] in [#10,#13] then begin
        LineBuf:=LineBuf+copy(Buf,StartPos,i-StartPos);
        if IsStdErr then
          TheErrors.Add(LineBuf)
        else
          TheOutput.Add(LineBuf);
        LineBuf:='';
        if (i<Count) and (Buf[i+1] in [#10,#13]) and (Buf[i]<>Buf[i+1])
        then
          inc(i);
        StartPos:=i+1;
      end;
      inc(i);
    end;
    LineBuf:=LineBuf+copy(Buf,StartPos,Count-StartPos+1);
  end;

var
  TheProcess: TProcess;
  OutputLine, StdErrLine: String;
  HasOutput: Boolean;
  ProcOpts: TProcessOptions;
  Cnt: Integer;
begin
  Result:=false;
  TheExitCode:=-1;
  if not FileIsExecutable(Executable) then
    begin
    debugln(['Error: [WebCore_RunTool] file is not executable "',Executable,'"']);
    if TheErrors<>nil then
      TheErrors.Add('WebCore_RunTool: file is not executable "'+Executable+'"');
    exit;
    end;

  SetLength(Buf{%H-},4096);
  try
    TheProcess := TProcess.Create(nil);
    try
      TheProcess.Executable:=Executable;
      TheProcess.Parameters.Assign(Params);
      ProcOpts:=[poUsePipes];
      if wcrtoConsole in Options then
        Include(ProcOpts,poNewConsole);
      TheProcess.Options:= ProcOpts;
      TheProcess.ShowWindow := swoHide;   // needed by Windows, ignored by Unix
      TheProcess.StartupOptions:=[suoUseShowWindow]; // needed by Windows, ignored by Unix
      TheProcess.CurrentDirectory:=UTF8ToSys(WorkingDirectory);
      TheProcess.Execute;
      OutputLine:='';
      StdErrLine:='';;
      SetLength(buf,4096);
      Cnt:=0;
      repeat
        HasOutput:=ReadInputPipe(TheProcess.Stderr,StdErrLine,true)
                or ReadInputPipe(TheProcess.Output,OutputLine,false);
        if (not HasOutput) then begin
          // no more pending output
          if not TheProcess.Running then break;
          // no more pending output and process is still running
          // => tool needs some time
          Sleep(50);
          if wcrtoNoWait in Options then begin
            inc(Cnt);
            if Cnt>20 then break;
          end;
        end;
      until false;
      // add rest of output
      if (OutputLine<>'') then
        TheOutput.Add(OutputLine);
      if (StdErrLine<>'') then
        TheErrors.Add(StdErrLine);
      if not (wcrtoNoWait in Options) then
        TheProcess.WaitOnExit;
      debugln(['Info: [TMSWebCoreIDEOptions.WebCore_RunTool] ExitCode=',TheProcess.ExitCode,' ExitStatus=',TheProcess.ExitStatus,' Errorlines=',TheErrors.Count]);
      TheExitCode:=TheProcess.ExitStatus;
      Result:=true;
    finally
      TheProcess.Free;
    end;
  except
    on E: Exception do begin
      debugln(['Error: [WebCore_RunTool] execute failed: "',Executable,'": '+E.Message]);
      if TheErrors<>nil then
        TheErrors.Add('WebCore_RunTool: executing "'+Executable+'" failed: '+E.Message);
    end;
  end;
end;

function WebCore_RunIDETool(const Executable, WorkingDirectory: string;
  Params: TStringList; ViewCaption, ViewName: string;
  const Options: TWCRunToolOptions): boolean;
var
  MsgResult, TheExitCode, i: Integer;
  OutLines, ErrLines: TStringList;
  RunResult: Boolean;
begin
  debugln(['Info: [WebCore_RunIDETool] ',ViewCaption,' Exe="',Executable,'" WorkDir="',WorkingDirectory,'"',
    ' Params=[',MergeCmdLineParams(Params),']']);
  Result:=false;
  repeat
    if not FileExistsUTF8(Executable) then
      begin
      debugln(['Warning: [WebCore_RunIDETool] file not found "',Executable,'"']);
      MsgResult:=IDEMessageDialog('Error',
        ViewCaption+':'+sLineBreak
        +'File not found: "'+Executable+'"'+sLineBreak
        +'Check Tools / Options / Environment / TMS Web Core / '+ViewName,
        mtError,[mbIgnore,mbRetry,mbCancel]);
      case MsgResult of
        mrIgnore: exit(true);
        mrRetry: continue;
      else exit(false);
      end;
      end;
    if FileIsExecutable(Executable) then
      break;
    debugln(['Warning: [WebCore_RunIDETool] file not executable "',Executable,'"']);
    MsgResult:=IDEMessageDialog('Error',
      ViewCaption+':'+sLineBreak
      +'Not executable: "'+Executable+'"'+sLineBreak
      +'Check Tools / Options / Environment / TMS Web Core / '+ViewName,
      mtError,[mbIgnore,mbRetry,mbCancel]);
    case MsgResult of
      mrIgnore: exit(true);
      mrRetry: continue;
    else exit(false);
    end;
  until false;

  OutLines:=TStringList.Create;
  ErrLines:=TStringList.Create;
  try
    debugln(['Info: [WebCore_RunIDETool] Params=[',MergeCmdLineParams(Params),']']);
    IDEMessagesWindow.AddCustomMessage(mluImportant,
      ViewCaption+': '+MaybeQuote(Executable)+' '+MergeCmdLineParams(Params),
      '',0,0,ViewCaption);

    {$IFDEF darwin}
    if wcrtoConsole in Options then
      RunResult:=WebCore_RunMacTerminal(Executable,WorkingDirectory,Params)
    else
    {$ENDIF}
    RunResult:=WebCore_RunTool(Executable,Params,WorkingDirectory,
                               OutLines,ErrLines,TheExitCode,Options)
        or (TheExitCode<>0);
    if ErrLines<>nil then
      for i:=0 to ErrLines.Count-1 do
      begin
        debugln(['Info: [WebCore_RunIDETool] ',ErrLines[i]]);
        IDEMessagesWindow.AddCustomMessage(mluImportant,ErrLines[i],'',0,0,ViewCaption);
      end;
    if not RunResult then
      begin
      IDEMessageDialog('Error',
        ViewCaption+':'+sLineBreak
        +'Unable to run "'+Executable+'"'+sLineBreak
        +'WorkingDir: "'+WorkingDirectory+'"'+sLineBreak
        +'Params:'+sLineBreak
        +Params.Text+sLineBreak
        +'ExitCode: '+IntToStr(TheExitCode),
        mtError,[mbOk]);
      exit(false);
      end;
    Result:=true;
  finally
    OutLines.Free;
    ErrLines.Free;
  end;
end;

function WebCore_RunMacTerminal(Executable, WorkingDirectory: string;
  Params: TStringList): boolean;
var
  OsaScript, Buffer: String;
  TheProcess: TProcess;
  InputStream: TStringStream;
  i: Integer;
  OutLen: LongInt;
begin
  Result:=false;
  OsaScript:=
    'tell app "Terminal"'+sLineBreak
    +'  set AppleScript''s text item delimiters to " "'+sLineBreak
    +'  do script '+AnsiQuotedStr(Executable,'"');
  for i:=0 to Params.Count-1 do
  begin
    if Pos(' ',Params[i])>0 then
      OsaScript:=OsaScript+' & " " & quoted form of ('+AnsiQuotedStr(Params[i],'"')+')'
    else
      OsaScript:=OsaScript+' & " '+Params[i]+'"';
  end;
  OsaScript:=OsaScript+sLineBreak
    +'end tell'+sLineBreak;

  InputStream:=TStringStream.Create(OsaScript);
  InputStream.Position:=0;
  TheProcess := TProcess.Create(nil);
  try
    TheProcess.Executable:='/usr/bin/osascript';
    TheProcess.Options:= [poUsePipes,poStderrToOutPut];
    TheProcess.CurrentDirectory:=UTF8ToSys(WorkingDirectory);
    debugln(['Info: [WebCore_RunMacTerminal] ',TheProcess.Executable,' Script={',OsaScript,'}']);
    TheProcess.Execute;
    TheProcess.Input.CopyFrom(InputStream,InputStream.Size-InputStream.Position);
    // read all output
    SetLength(Buffer{%H-},100);
    repeat
      if (TheProcess.Output=nil) then
        OutLen:=0
      else begin
        OutLen:=TheProcess.Output.NumBytesAvailable;
        if OutLen>0 then
          OutLen:=TheProcess.Output.Read(Buffer[1],Min(length(Buffer),OutLen));
      end;
      if OutLen=0 then break;
      debugln(['Info: [WebCore_RunMacTerminal] ',Buffer]);
    until false;
    //debugln(['Info: [WebCore_RunMacTerminal] output end']);

    // do not: TheProcess.WaitOnExit;
    debugln(['Info: [TMSWebCoreIDEOptions.WebCore_RunMacTerminal] ExitCode=',TheProcess.ExitCode,' ExitStatus=',TheProcess.ExitStatus]);
    Result:=true;
  finally
    TheProcess.Free;
    InputStream.Free;
  end;
end;

function TWebCoreEventSink.ExecWebServer(Prj: TLazProject): boolean;
Var
  Executable, WorkDir, CmdLineParams: String;
  Params: TStringListUTF8;
  i: Integer;
  StartWebServer, ParamsChanged, IsRunning: Boolean;
  ProcOpts: TWCRunToolOptions;
begin
  Result:=false;
  if Prj=nil then exit;
  StartWebServer:=WebCoreDefaultStartWebServer;
  if Prj.CustomData.Contains(KeyStartWebServer) then
    StartWebServer:=Prj.CustomData[KeyStartWebServer]='1';
  if not StartWebServer then
    exit(true);

  Executable:=WebCoreOptions.GetParsedOptionValue(wccoWebServer);
  WorkDir:=ExtractFilePath(Executable);
  CmdLineParams:=WebCoreOptions.GetParsedOptionValue(wccoWebServerParams);

  ParamsChanged:=(Executable<>FLastServerExecutable)
      or (WorkDir<>FLastServerWorkDir)
      or (CmdLineParams<>FLastServerCmdLineParams);
  IsRunning:=IsWebServerRunning(Prj,200);

  if ParamsChanged or not IsRunning then
    begin
    if ParamsChanged and IsRunning then
      debugln(['Info: (lazarus) [TWebCoreEventSink.ExecWebServer] server answered, but params have changed -> start again']);
    FLastServerExecutable:=Executable;
    FLastServerWorkDir:=WorkDir;
    FLastServerCmdLineParams:=CmdLineParams;
    Params:=TStringListUTF8.Create;
    try
      // run webserver manager
      SplitCmdLineParams(CmdLineParams,Params);
      ProcOpts:=[wcrtoConsole];
      if not WebCore_RunIDETool(Executable,WorkDir,Params,
          WebCoreWebServerCaption,'Web Server',
          ProcOpts) then
        exit;
    finally
      Params.Free;
    end;
    // give it some seconds to start
    IsRunning:=false;
    for i:=1 to 10 do
      begin
      Application.ProcessMessages;
      if IsWebServerRunning(Prj,200) then
      begin
        IsRunning:=true;
        break;
      end;
      Sleep(100);
      end;
    end;
  if not IsRunning then
  begin
    IDEMessagesWindow.AddCustomMessage(mluWarning,
      WebCoreWebServerCaption+': failed HTTP on '+MaybeQuote(WebCoreOptions.GetParsedOptionValue(wccoURL)),
      '',0,0,WebCoreWebServerCaption);
  end;
  Result:=true;
end;

function TWebCoreEventSink.ExecWebRunner(Prj: TLazProject): boolean;
Var
  Executable, WorkDir, aCaption, aName, IniFile, URL: String;
  Params: TStringListUTF8;
begin
  Result:=false;
  //debugln(['TWebCoreEventSink.ExecWebRunner START']);
  if Prj=nil then exit;
  if not UseWebRunner(Prj) then
    exit(true);
  Params:=TStringListUTF8.Create;
  try
    // run webrunner
    //debugln(['TWebCoreEventSink.ExecWebRunner WebCoreOptions.WebRunner=',WebCoreOptions.WebRunner]);
    if WebCoreOptions.WebRunner<>'' then
    begin
      // run webrunner to start the browser
      Executable:=WebCoreOptions.GetParsedOptionValue(wccoWebRunner);
      if Executable='' then
      begin
        IDEMessageDialog('Cannot find webrunner',
          'See Tools / Options / TMS Web Core / WebRunner',mtError,[mbOk]);
        exit(false);
      end;
      IniFile:=WebCoreOptions.GetWebRunnerIniFilename(Prj);
      if (Executable='') or not FileExistsUTF8(IniFile) then
      begin
        IDEMessageDialog('Missing webrunner ini',
          'Ini="'+IniFile+'"'+#13
          +'[20200618170904]',mtError,[mbOk]);
        exit(false);
      end;
      Params.Add(IniFile);
      aName:='Web Runner';
      aCaption:='Web Runner';
    end else begin
      // start browser
      URL:=WebCoreOptions.GetParsedOptionValue(wccoURL);
      if URL='' then
        debugln(['Warn: [TWebCoreEventSink.ExecWebRunner] URL is empty']);
      Executable:=WebCoreOptions.GetBrowserPath(URL,Params);
      if Executable='' then
      begin
        if WebCoreOptions.Browser in [bNothing,bDefault,bSystem] then
        begin
          debugln(['Hint: [TWebCoreEventSink.ExecWebRunner] using fallback OpenURL(',URL,')']);
          Result:=OpenURL(URL);
          exit;
        end;
      end;
      if Executable='' then
      begin
        IDEMessageDialog('Cannot find browser',
          'See Tools / Options / TMS Web Core / Browser',mtError,[mbOk]);
        exit(false);
      end;
      aName:='Browser';
      aCaption:='Browser('+ExtractFileName(Executable)+')';

    end;
    // work dir
    WorkDir:=Prj.LazCompilerOptions.GetUnitOutputDirectory(false);
    //debugln(['TWebCoreEventSink.ExecWebRunner Exe="',Executable,'" aName="',aName,'" aCaption="',aCaption,'" WorkDir="',WorkDir,'"']);
    if WorkDir='' then
      WorkDir:=ExtractFilePath(Executable);
    // execute
    if not WebCore_RunIDETool(Executable,WorkDir,Params,aCaption,aName,[wcrtoNoWait]) then
      exit;
    Result:=true;
  finally
    Params.Free;
  end;
end;

function TWebCoreEventSink.CreateDummyGDBExe(Prj: TLazProject): boolean;
{$IFDEF Windows}
const
  LibName = 'libiconv-2.dll';
var
  LazarusDir, TargetOS, TargetCPU, SrcGDB, DestGDB, Dir, MingwDir,
    DestLib, SrcLib: String;
  CompOpts: TLazCompilerOptions;
{$ENDIF}
begin
  Result:=true;
  if Prj=nil then exit;
  {$IF defined(Windows)}
  // the Lazarus windows installer sets debugger path to
  // lazarusdir\mingw\$(TargetCPU)-$(TargetOS)\bin\gdb.exe
  LazarusDir:='$(LazarusDir)';
  IDEMacros.SubstituteMacros(LazarusDir);
  if LazarusDir='' then
  begin
    debugln(['Warning: [TWebCoreEventSink.CreateDummyGDBExe] LazarusDir=""']);
    exit(true); // does not make sense -> skip
  end;
  LazarusDir:=AppendPathDelim(LazarusDir);
  MingwDir:=LazarusDir+'mingw';
  if not DirPathExists(MingwDir) then
  begin
    debugln(['Info: [TWebCoreEventSink.CreateDummyGDBExe] dir missing: ',MingwDir]);
    exit(true); // this Lazarus does not have a preinstalled mingw -> ok
  end;

  // check if dummy gdb already exists
  CompOpts:=Prj.LazCompilerOptions;
  TargetOS:=lowercase(CompOpts.TargetOS);
  if TargetOS='' then
    TargetOS:='browser';
  TargetCPU:=lowercase(CompOpts.TargetCPU);
  if TargetCPU='' then
    TargetCPU:='ecmascript5';
  DestGDB:=MingwDir+'\'+TargetCPU+'-'+TargetOS+'\bin\gdb.exe';
  DestLib:=ExtractFilePath(DestGDB)+LibName;
  if FileExistsCached(DestGDB) and FileExistsCached(DestLib) then
  begin
    debugln(['Info: [TWebCoreEventSink.CreateDummyGDBExe] dummy gdb already exists: ',DestGDB]);
    exit(true); // dummy gdb already exists -> ok
  end;

  // find a preinstalled gdb.exe
  SrcGDB:=MingwDir+'\x86_64-win64\bin\gdb.exe';
  if not FileExistsCached(SrcGDB) then
  begin
    debugln(['Info: [TWebCoreEventSink.CreateDummyGDBExe] 64bit gdb.exe not found: ',SrcGDB]);
    SrcGDB:=MingwDir+'\i386-win32\bin\gdb.exe';
    if not FileExistsCached(SrcGDB) then
    begin
      debugln(['Info: [TWebCoreEventSink.CreateDummyGDBExe] 32bit gdb.exe not found: ',SrcGDB]);
      exit(true); // not found
    end;
  end;
  SrcLib:=ExtractFilePath(SrcGDB)+LibName;

  Dir:=ChompPathDelim(ExtractFilePath(DestGDB));
  if not ForceDirectoriesUTF8(Dir) then
  begin
    debugln(['Warning: [TWebCoreEventSink.CreateDummyGDBExe] ForceDirectoriesUTF8 failed: ',Dir]);
    if IDEMessageDialog('Error',
        'Unable to create directory "'+Dir+'".',
        mtError,[mbCancel,mbIgnore])=mrIgnore then
      exit(true);
    exit(false)
  end;
  if not FileExistsCached(DestGDB)
      and not CopyFile(SrcGDB,DestGDB,false,false) then
  begin
    debugln(['Warning: [TWebCoreEventSink.CreateDummyGDBExe] CopyFile failed: "',SrcGDB,'" -> "',DestGDB,'"']);
    if IDEMessageDialog('Error',
        'Unable to copy file from'+sLineBreak
        +SrcGDB+sLineBreak
        +'to'+sLineBreak
        +DestGDB,
        mtError,[mbCancel,mbIgnore])=mrIgnore then
      exit(true);
    exit(false)
  end;
  if FileExistsCached(SrcLib) and not FileExistsCached(DestLib) then
  begin
    // Lazarus 2.0.8 and below needed a libiconv.dll
    if CopyFile(SrcLib,DestLib,false,false) then
      InvalidateFileStateCache
    else begin
      debugln(['Warning: [TWebCoreEventSink.CreateDummyGDBExe] CopyFile failed: "',SrcLib,'" -> "',DestLib,'"']);
      if IDEMessageDialog('Error',
          'Unable to copy file from'+sLineBreak
          +SrcLib+sLineBreak
          +'to'+sLineBreak
          +DestLib,
          mtError,[mbCancel,mbIgnore])=mrIgnore then
        exit(true);
      exit(false)
    end;
  end;
  {$ENDIF}
end;

function TWebCoreEventSink.UseWebRunner(Prj: TLazProject): boolean;
begin
  if Prj.CustomData.Contains(KeyRunWebRunner) then
    Result:=Prj.CustomData[KeyRunWebRunner]='1'
  else
    Result:=WebCoreDefaultRunWebRunner;
end;

function TWebCoreEventSink.CheckLicense(Prj: TLazProject): boolean;
// check existence of a license and try to help user installing a license
var
  WebCompiler: string;

  function RunLicenseGenerator(WorkDir, Generator, LicensePos: string): boolean;
  var
    Params, TheOutput: TStringListUTF8;
    TheExitCode, r: integer;
  begin
    if FileExistsUTF8(Generator) then
      r:=IDEQuestionDialog('TMSWebcompiler license not found',
        LicensePos
          +'This license is generated using your email address and TMS Webcore key.',
        mtWarning,[mrIgnore,mrYesToAll,'Ignore until IDE restarts',mrRetry,'Run license generator','isdefault',mrCancel])
    else
      r:=IDEQuestionDialog('TMSWebcompiler license not found',
        LicensePos
          +'There is no license generator at'+LineEnding
          +'"'+Generator+'".',
        mtWarning,[mrIgnore,mrYesToAll,'Ignore until IDE restarts',mrCancel]);
    case r of
    mrIgnore: exit(true);
    mrYesToAll:
      begin
        WebCoreOptions.CheckLicense:=false;
        exit(true);
      end;
    mrRetry:
      begin
        // run license generator
        Params:=TStringListUTF8.Create;
        TheOutput:=TStringListUTF8.Create;
        try
          WebCore_RunTool(Generator,Params,WorkDir,TheOutput,nil,TheExitCode,[wcrtoNoWait]);
          exit(false);
        finally
          TheOutput.Free;
          Params.Free;
        end;
      end;
    else
      exit(false);
    end;
  end;

  function CheckUnixLicense: boolean;
  var
    LicenseFile, Generator, WorkDir: String;
  begin
    WorkDir:=ChompPathDelim(ExtractFilePath(WebCompiler));
    LicenseFile:=WorkDir+'/License.lic';
    if FileExistsUTF8(LicenseFile) then
      exit(true);
    {$IFDEF Darwin}
    Generator:=WorkDir+'/TMSLicenseGenerator.app/Contents/MacOS/TMSLicenseGenerator';
    {$ELSE}
    Generator:=WorkDir+'/TMSLicenseGenerator_Linux';
    {$ENDIF}
    Result:=RunLicenseGenerator(WorkDir,Generator,
      'The TMSWebCompiler requires a license file'+LineEnding
      +'"'+LicenseFile+'".'+LineEnding);
  end;

  {$IFDEF Windows}
  function WinRegistryLicenseExists: boolean;
  // false if license is missing
  var
    r: TRegistry;
    s: String;
  begin
    try
      r:=TRegistry.Create;
      try
        r.RootKey:=HKEY_CURRENT_USER;
        if not r.OpenKeyReadOnly('\SOFTWARE\tmssoftware\TMS WEB Core') then
          exit(false);
        s:=r.ReadString('License');
        if s='' then
          exit(false);
        if length(s)<10 then
          exit(false);
      finally
        r.Free;
      end;
    except
      // exception: unable to check license
    end;
    Result:=true;
  end;
  {$ENDIF}

  function CheckWinLicense: boolean;
  var
    WorkDir, Generator, LicenseFile: String;
  begin
    {$IFDEF Windows}
    // check registry
    if WinRegistryLicenseExists then
      exit(true);
    {$ENDIF}
    // check license file
    WorkDir:=ChompPathDelim(ExtractFilePath(WebCompiler));
    LicenseFile:=WorkDir+'/License.lic';
    if FileExistsUTF8(LicenseFile) then
      exit(true);

    WorkDir:=WebCoreOptions.GetParsedOptionValue(wccoTMSWebDir)+'\Bin\Win32';
    Generator:=WorkDir+'\TMSLicenseGenerator.exe';
    Result:=RunLicenseGenerator(WorkDir,Generator,
      'The TMSWebCompiler requires a license either in the registry string'+LineEnding
      +'"HKEY_CURRENT_USER\SOFTWARE\tmssoftware\TMS WEB Core\License"'+LineEnding
      +'or in the file'+LineEnding
      +'"'+LicenseFile+'"'+LineEnding);
  end;

begin
  if Prj=nil then ;
  if not WebCoreOptions.CheckLicense then
    exit(true);
  WebCompiler:=WebCoreOptions.GetParsedOptionValue(wccoWebCompiler);
  if WebCompiler='' then exit(true);
  if not FileExistsCached(WebCompiler) then
    exit(true); // no tmswebcompiler -> will raise another error, cannot check license
  {$IFDEF Unix}
  Result:=CheckUnixLicense;
  {$ELSE}
    {$IFDEF Windows}
    Result:=CheckWinLicense;
    {$ELSE}
    Result:=true;
    {$ENDIF}
  {$ENDIF}
end;

function TWebCoreEventSink.CreateWebRunnerIni(aFileName: String): boolean;
const
  SPaths = 'Paths';
  KeyHTMLPath = 'HtmlPath';
  KeyHTMLFile = 'HtmlFile';
  KeyDefaultURL = 'DefaultURL';
  KeyURL = 'URL';
  KeySingleInstance = 'SingleInstance';
  KeyDebug = 'Debug';
  KeyBrowser = 'Browser';
  KeyElectron = 'Electron';
  KeyElectronBuild = 'ElectronBuild';
  KeyElectronNPM = 'ElectronNPM';
Var
  Prj: TLazProject;
  NewStream, OldStream: TMemoryStream;
  Ini: TMemIniFile;
  MsgResult: Integer;
  SrcEdit: TSourceEditorInterface;
  HTMLPath, HTMLFile, URL, DefaultURL: String;
begin
  Result:=false;
  Prj:=LazarusIDE.ActiveProject;
  if Prj=nil then exit;

  if AFileName='' then
  begin
    if WebCoreOptions.WebRunner='' then
      exit(true); // no web runner -> do no create ini
    AFileName:=WebCoreOptions.GetWebRunnerIniFilename(Prj);
  end;
  DebugLn(['Hint: [TWebCoreEventSink.CreateWebRunnerIni]: Creating "',aFileName,'"']);
  OldStream:=nil;
  NewStream:=TMemoryStream.Create;
  try
    if FileExistsUTF8(aFileName) then
      begin
      OldStream:=TMemoryStream.Create;
      repeat
        try
          OldStream.LoadFromFile(aFileName);
          MsgResult:=mrOk;
        except
          on E: Exception do begin
            MsgResult:=IDEMessageDialog('Read error',
              'Unable to read file "'+aFileName+'":'+sLineBreak
              +E.Message,mtError,[mbIgnore,mbRetry,mbCancel]);
          end;
        end;
        case MsgResult of
          mrCancel: exit(false);
          mrIgnore: exit(true);
          mrRetry: continue;
        else break;
        end;
      until false;
      OldStream.Position:=0;
      NewStream.CopyFrom(OldStream,OldStream.Size);
      NewStream.Position:=0;
      end;

    HTMLPath:=ChompPathDelim(WebCoreOptions.GetOutputDir(Prj));
    HTMLFile:=ExtractFileName(WebCoreOptions.GetHTMLFilename(Prj));

    if not DirPathExists(HTMLPath) then
      repeat
        debugln(['Info: TWebCoreEventSink.CreateWebRunnerIni create directory: ',HTMLPath]);
        if ForceDirectoriesUTF8(ChompPathDelim(HTMLPath)) then break;
        MsgResult:=IDEMessageDialog('Error',
          'Unable to create directory "'+HTMLPath+'"',
          mtError,[mbRetry,mbCancel]);
        case MsgResult of
          mrRetry: ;
        else
          exit(false);
        end;
      until false;

    DefaultURL:=WebCoreOptions.GetDefaultURL;
    URL:=DefaultURL;

    Ini:=TMemIniFile.Create(NewStream);
    Ini.CacheUpdates:=true;
    Ini.WriteString(SPaths,KeyHtmlPath,HTMLPath);
    Ini.WriteString(SPaths,KeyHtmlFile,HTMLFile);
    Ini.WriteString(SPaths,KeyDefaultURL,DefaultURL);
    Ini.WriteBool(SPaths,KeySingleInstance,False);
    Ini.WriteBool(SPaths,KeyDebug,False);
    Ini.WriteString(SPaths,KeyURL,URL);
    Ini.WriteInteger(SPaths,KeyBrowser,BrowserToWebRunnerID[WebCoreOptions.Browser]);
    Ini.WriteBool(SPaths,KeyElectron,False);
    Ini.WriteBool(SPaths,KeyElectronBuild,False);
    Ini.WriteString(SPaths,KeyElectronNPM,'');
    // write ini
    Ini.UpdateFile;
    if (OldStream=nil)
        or (OldStream.Size<>NewStream.Size)
        or ((OldStream.Size>0) and not CompareMem(OldStream.Memory,NewStream.Memory,OldStream.Size))
        then
      begin
      // save tmswebrunner.ini
      repeat
        try
          NewStream.Position:=0;
          NewStream.SaveToFile(aFileName);
          MsgResult:=mrOK;
        except
          on E: Exception do begin
            MsgResult:=IDEMessageDialog('Write error',
              'Unable to write file "'+aFileName+'":'+sLineBreak
              +E.Message,mtError,[mbIgnore,mbRetry,mbCancel]);
          end;
        end;
        case MsgResult of
          mrCancel: exit(false);
          mrIgnore: exit(true);
          mrRetry: continue;
        else break;
        end;
      until false;
      // revert editor changes
      SrcEdit:=SourceEditorManagerIntf.SourceEditorIntfWithFilename(aFileName);
      if SrcEdit<>nil then
        LazarusIDE.DoOpenEditorFile(aFileName,-1,-1,[ofRevert,ofRegularFile]);
      end;
  finally
    OldStream.Free;
    Ini.Free;
    NewStream.Free;
  end;
  Result:=true;
end;

procedure TWebCoreEventSink.DoHeaders(Sender: TObject);
var
  i: Integer;
begin
  debugln(['Info: [TWebCoreEventSink.DoHeaders] Response headers received START:']);
  With (Sender as TFPHTTPClient) do
    For i:=0 to ResponseHeaders.Count-1 do
      debugln('  '+ResponseHeaders[i]);
  debugln(['Info: [TWebCoreEventSink.DoHeaders] Response headers received END']);
end;

function TWebCoreEventSink.IsWebServerRunning(Prj: TLazProject;
  TimeoutInMS: integer): boolean;
var
  URL, Response: String;
begin
  Result:=false;
  if Prj=nil then ;
  URL:=WebCoreOptions.GetParsedOptionValue(wccoURL);
  if (URL<>'') and (URL[1] in ['''','"']) then
    URL:=AnsiDequotedStr(URL,URL[1]);
  if URL='' then
    begin
    URL:=WebCoreDefault_URL;
    IDEMacros.SubstituteMacros(URL);
    end;
  With TFPHTTPClient.Create(Nil) do
    try
      AllowRedirect:=True;
      //OnRedirect:=@ShowRedirect;
      //OnPassword:=@DoPassword;
      //OnDataReceived:=@DoProgress;
      OnHeaders:=@DoHeaders;
      IOTimeout:=TimeoutInMS;
      try
        debugln(['Info: [TWebCoreEventSink.IsWebServerRunning] URL="',URL,'"']);
        Response:=Get(URL);
        debugln(['Info: [TWebCoreEventSink.IsWebServerRunning] Response="',Response,'"']);
        Result:=Response<>'';
      except
        on E: Exception do
          debugln(['Info: [TWebCoreEventSink.IsWebServerRunning] Error: "',E.Message,'"']);
      end;
    finally
      Free;
    end;
end;

function TWebCoreEventSink.ProjectBuilding(Sender: TObject): TModalResult;
var
  Prj: TLazProject;
  CompOpts: TLazCompilerOptions;
  HTMLDir, LPIFilename, ProjDir, NewTargetFilename: String;
  PkgList: TFPList;
  Pkg: TIDEPackage;
  MsgResult, i: Integer;
begin
  Result:=mrOK;
  Prj:=LazarusIDE.ActiveProject;
  if Prj=nil then exit;
  if (Prj.CustomData[KeyTMSWebCoreProject]<>'1') then
    exit;
  DebugLn('Info: [TWebCoreEventSink.ProjectBuilding] called');

  LPIFilename:=Prj.ProjectInfoFile;

  // check if project was saved, tmswebcompiler needs the lpi
  if not FilenameIsAbsolute(LPIFilename) then
    begin
    IDEMessageDialog('Error',
      'Compiling a TMS WebCore project in the temporary directory is not supported yet.'+sLineBreak
      +'Please save the project first!',
      mtError,[mbOk]);
    exit(mrCancel);
    end;

  ProjDir:=WebCoreOptions.GetProjectDir(Prj);
  CompOpts:=Prj.LazCompilerOptions;
  HTMLDir:=CompOpts.GetUnitOutputDirectory(false);
  HTMLDir:=CreateRelativePath(HTMLDir,ProjDir);
  NewTargetFilename:=AppendPathDelim(HTMLDir)
                        +ChangeFileExt(ExtractFileName(LPIFilename),'.js');
  if CompOpts.TargetFilename<>NewTargetFilename then
    begin
    debugln(['Info: [TWebCoreEventSink.ProjectBuilding] updating TargetFilename to "',NewTargetFilename,'"']);
    CompOpts.TargetFilename:=NewTargetFilename;
    end;

  // check if user accidentally added LCL
  PkgList:=nil;
  try
    PackageEditingInterface.GetRequiredPackages(Prj,PkgList,[pirSkipDesignTimeOnly]);
    if PkgList<>nil then
      for i:=0 to PkgList.Count-1 do
        begin
        Pkg:=TIDEPackage(PkgList[i]);
        case lowercase(Pkg.Name) of
          'lcl','lclbase','lazutils','fcl':
            begin
            MsgResult:=IDEMessageDialog('Error',
              'A WebCore project cannot be combined with package '+Pkg.Name+sLineBreak
              +'Check Project Inspector''s required packages.',
              mtError,[mbIgnore,mbCancel]);
            case MsgResult of
              mrIgnore: break;
            else
              exit(mrCancel);
            end;
            end;
        end;
        end;
  finally
    PkgList.Free;
  end;

  // create webrunner.ini
  if not CreateWebRunnerIni then
    exit(mrCancel);

  if not CreateDummyGDBExe(Prj) then
    exit(mrCancel);
end;

function TWebCoreEventSink.RunDebugInit(Sender: TObject; var Handled: boolean
  ): TModalResult;
var
  Prj: TLazProject;
begin
  Result:=mrOk;
  Prj:=LazarusIDE.ActiveProject;
  if Prj=nil then exit;
  debugln(['Info: [TWebCoreEventSink.RunDebug]: Prj.CustomData[KeyTMSWebCoreProject]=',Prj.CustomData[KeyTMSWebCoreProject]]);
  if (Prj.CustomData[KeyTMSWebCoreProject]<>'1') then
    exit;
  Handled:=true;
  try
    if not CheckLicense(Prj) then
      exit(mrCancel);
    {$IF Laz_FullVersion>=2010000}
    // after 2.1.0  it is called before building (using AddHandlerOnRunDebugInit)
    Result:=LazarusIDE.DoBuildProject(crRun,[]);
    if Result<>mrOk then begin
      debugln(['Info: [TWebCoreEventSink.RunDebugInit] DoBuildProject failed']);
      exit;
    end;
    {$ELSE}
    // before 2.1.0 it is called after building (using AddHandlerOnRunDebug)
    {$ENDIF}
    // start webserver
    if not ExecWebServer(Prj) then
      exit(mrCancel);
    // start browser
    if not ExecWebRunner(Prj) then
      exit(mrCancel);
    debugln(['Info: [TWebCoreEventSink.RunDebugInit] success']);
  finally
    LazarusIDE.ToolStatus:=itNone;
  end;
end;

function TWebCoreEventSink.RunWithoutDebug(Sender: TObject; var Handled: boolean
  ): TModalResult;
var
  Prj: TLazProject;
begin
  Result:=mrOk;
  Prj:=LazarusIDE.ActiveProject;
  if Prj=nil then exit;
  debugln(['Info: [TWebCoreEventSink.RunWithoutDebug]: Prj.CustomData[KeyTMSWebCoreProject]=',Prj.CustomData[KeyTMSWebCoreProject]]);
  if (Prj.CustomData[KeyTMSWebCoreProject]<>'1') then
    exit;
  try
    if not CheckLicense(Prj) then
      exit(mrCancel);
    if not ExecWebServer(Prj) then
      exit(mrCancel);
    if UseWebRunner(Prj) then
      begin
      Handled:=true;
      if not ExecWebRunner(Prj) then
        exit(mrCancel);
      end
    else
      begin
      // if not using WebRunner continue with whatever the user set in Run parameters
      end;
    debugln(['Info: [TWebCoreEventSink.RunDebug] webserver started successfully']);
  finally
    LazarusIDE.ToolStatus:=itNone;
  end;
end;

function TWebCoreEventSink.SaveEditorFile(Sender: TObject;
  aFile: TLazProjectFile; SaveStep: TSaveEditorFileStep; TargetFilename: string
  ): TModalResult;
var
  Prj: TLazProject;
  aFilename: String;
begin
  Result:=mrOk;
  Prj:=LazarusIDE.ActiveProject;
  if Prj=nil then exit;
  if (Prj.CustomData[KeyTMSWebCoreProject]<>'1') then
    exit;
  aFilename:=aFile.Filename;
  //debugln(['Info: [TWebCoreEventSink.SaveEditorFile] ',aFilename,' IsMain=',aFile=Prj.MainFile,' After=',SaveStep=sefsAfterWrite]);
  if aFile=Prj.MainFile then
    begin
    // saving main file
    if SaveStep<>sefsAfterWrite then exit;
    debugln(['Info: [TWebCoreEventSink.SaveEditorFile] after saving main file ',aFilename]);
    end;
  if Sender=nil then ;
  if TargetFilename='' then ;
end;

function TWebCoreEventSink.DesignerAddClicked(ADesigner: TIDesigner;
  MouseDownComponent: TComponent; Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer; var AComponentClass: TComponentClass; var NewParent: TComponent
  ): boolean;
var
  Prj: TLazProject;
begin
  Result:=true;
  Prj:=LazarusIDE.ActiveProject;
  if Prj=nil then exit;
  if (Prj.CustomData[KeyTMSWebCoreProject]<>'1') then
    exit;
  if ADesigner.LookupRoot is WebFormClass then
    begin
    // this is a webform -> only allow specific components
    Result:=WebCoreOptions.IsComponentAllowed(AComponentClass);
    if not Result then
      IDEMessageDialog('Error',
        WebFormClass.ClassName+' does not support "'+AComponentClass.ClassName+'".',
        mtError,[mbCancel]);
    end;
end;

Procedure DoneWebCoreOptions;
begin
  FreeAndNil(EventSink);
  if WebCoreOptions<>nil then
  begin
    try
      if WebCoreOptions.Modified then
        WebCoreOptions.Save;
    except
    end;
    FreeAndNil(WebCoreOptions);
  end;
end;

finalization
  DoneWebCoreOptions;
end.
